/* 
    Argonne Leadership Computing Facility benchmark
    BlueGene/P version
    Collectives over sliced subcommunicator delivered latency
    Written by Vitali Morozov <morozov@anl.gov>

    Measures broadcast over a user-defined subcommunicator in vn mode

    The subcommunicator is: 
        tasks on CPU 0 belong to one group
        tasks on CPU 1 belong to the second group
        tasks on CPU 2 belong to the third group
        tasks on CPU 3 belong to the fourth group

        tasks on node 0 broadcast a 15MB buffer to other tasks within a group
        tests DMA on node 0
*/    
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define MAXN 250              // repetition rate for a single pair test
#define MAXSKIP 10              // skip first tests to warm-up

#define BSIZE 15*1024*1024*8     // largest broadcast message size, bytes

unsigned long long getticks();
int split( MPI_Comm *Split_comm );

static MPI_Comm split_comm;

main( int argc, char *argv[] )
{
    int taskid, ntasks, k, split_id, myt;
    unsigned char buf[ BSIZE+2 ];
    double d, dtmin, dtmax;
    unsigned long long t1;
    
    MPI_Init( &argc, &argv );
    MPI_Comm_rank( MPI_COMM_WORLD, &taskid );
    
    myt = split( &split_comm );
    MPI_Comm_rank( split_comm, &split_id );
    
    if ( split_id == 0 )
        buf[0] = (unsigned char)(myt);  // must be 0, 1, 2, or 3
    
    MPI_Barrier( MPI_COMM_WORLD );
    
    for ( k = 0; k < MAXSKIP; k++ )
    {
        if ( split_id == 1 )
            MPI_Bcast( (void *)&(buf[1]), BSIZE, MPI_BYTE, 0, split_comm ); 
        else    
            MPI_Bcast( (void *)&(buf[0]), BSIZE, MPI_BYTE, 0, split_comm ); 
    }
    
    t1 = getticks();
    for ( k = 0; k < MAXN; k++ )
    {
        if ( split_id == 1 )
            MPI_Bcast( (void *)&(buf[1]), BSIZE, MPI_BYTE, 0, split_comm ); 
        else    
            MPI_Bcast( (void *)&(buf[0]), BSIZE, MPI_BYTE, 0, split_comm ); 
    }
    t1 = getticks() - t1;
    
    d = (double)t1 / (double)MAXN / (double)850.;  /* in us */
    
    MPI_Reduce( &d, &dtmin, 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, MPI_COMM_WORLD );
    MPI_Reduce( &d, &dtmax, 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, MPI_COMM_WORLD );

    if ( taskid == 0 )
        fprintf( stderr, "Broadcast %d bytes, min, max, s: %18.6lf   %18.6lf\n", BSIZE, dtmin*1e-6, dtmax*1e-6 );

    fprintf( stderr, "%d: buf[0] = %d\n", taskid, (int)(buf[0]) );

    MPI_Finalize();
    return 0;
}
