#include <mpi.h>
#include <stdio.h>
#include <common/bgp_personality.h>
#include <common/bgp_personality_inlines.h>
#include <spi/kernel_interface.h> 

/* Returns the source task and the list of the nearest target tasks (1 hop distance)
   For partition (X,Y,Z), 
      the source is the task, placed to (x/2,y/2,z/2,t) to guarantee the maximal connectivity
      t is the smallest available
   The target[N] list is x+-1,y,z,t or x,y+-1,z,t or x,y,z+-1,t
    
   Returns:
        0, if source and target are assigned
        1, if source node does not contain tasks
        2, if one of the target nodes does not contain tasks
        3, if N != 6, BG/P 3D mesh/torus
*/    
int getranks( int N, int *source, int *target )
{
  _BGP_Personality_t personality;
  _BGP_Personality_t *p = &personality;
  unsigned int x, y, z, t, size, rank, i;
 
  if ( N != 6 ) return 3;

  /* determine the source node */
  Kernel_GetPersonality(p, sizeof(personality));

  x = BGP_Personality_xSize( p ) / 2;
  y = BGP_Personality_ySize( p ) / 2;
  z = BGP_Personality_zSize( p ) / 2;
  
  t = Kernel_PhysicalProcessorID();
  if ( ( Kernel_ProcessCount() == 2 ) && ( t == 2 ) ) t = 1;

  /* assign taskid to the source node */
  Kernel_Coord2Rank( x, y, z, t, (unsigned *)source, &size );

  if ( (*source < 0) || (*source >= size) )
      return 1;

  /* assign taskids to the targets */
  /* x+1 */
  Kernel_Coord2Rank( x+1, y, z, t, (unsigned *)&target[0], &size );
  if ( (target[0] < 0) || (target[0] >= size) ) return 2;
  
  /* x-1 */
  Kernel_Coord2Rank( x-1, y, z, t, (unsigned *)&target[1], &size );
  if ( (target[1] < 0) || (target[1] >= size) ) return 2;

  /* y+1 */
  Kernel_Coord2Rank( x, y+1, z, t, (unsigned *)&target[2], &size );
  if ( (target[2] < 0) || (target[2] >= size) ) return 2;

  /* y-1 */
  Kernel_Coord2Rank( x, y-1, z, t, (unsigned *)&target[3], &size );
  if ( (target[3] < 0) || (target[3] >= size) ) return 2;

  /* z+1 */
  Kernel_Coord2Rank( x, y, z+1, t, (unsigned *)&target[4], &size );
  if ( (target[4] < 0) || (target[4] >= size) ) return 2;

  /* z-1 */
  Kernel_Coord2Rank( x, y, z-1, t, (unsigned *)&target[5], &size );
  if ( (target[5] < 0) || (target[5] >= size) ) return 2;

  return 0;
}
