#include <mpi.h>
#include <stdio.h>
#include <common/bgp_personality.h>
#include <common/bgp_personality_inlines.h>
#include <spi/kernel_interface.h> 

/* Returns two communication tasks, participating in bi-section test
   1) Determine the partition size (X,Y,Z)
   2) Find the largest dimension S. This dimension will cut the partition in two halves
   3) Determine my coodinates (x,y,z,t) 
   
   a) S = X. Participating sending tasks (0,y,z,t) will communicate with receiveng tasks (1,y,z,t), other tasks are ignored.   
   b) S = Y. Participating sending tasks (x,0,z,t) will communicate with receiveng tasks (x,1,z,t), other tasks are ignored.   
   c) S = Y. Participating sending tasks (x,y,0,t) will communicate with receiveng tasks (x,y,1,t), other tasks are ignored.   
        
   Returns:
        0, if source and target are assigned for participating tasks or set -1 to nonparticipating tasks
        1, if cannot assign a target, nothing runs on the core
        1, if cannot assign a source, nothing runs on the core
*/    
int getranks( int rank, int *source, int *target )
{
  _BGP_Personality_t personality;
  _BGP_Personality_t *p = &personality;
  int J, isTorus;
  uint32_t P[3], c[4], sc[4], rc[4], size;

  /* determine the partition size */
  Kernel_GetPersonality(p, sizeof(personality));
  
  P[0] = BGP_Personality_xSize( p );
  P[1] = BGP_Personality_ySize( p );
  P[2] = BGP_Personality_zSize( p );
  
  /* find the longest dimension, let it be J */ 
  J = 0;
  if ( P[J] < P[1] ) J = 1;
  if ( P[J] < P[2] ) J = 2;

  /* find if this dimension is a torus */
  if( J == 0 ) isTorus = BGP_Personality_isTorusX( p );
  if( J == 1 ) isTorus = BGP_Personality_isTorusY( p );
  if( J == 2 ) isTorus = BGP_Personality_isTorusZ( p );

  /* determine my coordinates */
  Kernel_Rank2Coord( rank, &c[0], &c[1], &c[2], &c[3] );

  // for mesh  in J dimension, 1 sender and 1 receiver: 0 <-> 1
  // for torus in J dimension, 2 pairs: 0 <-> 1 and P[J]/2 <-> P[J]/2+1
  
  // for non-participating tasks
  *source = -1;
  *target = -1;

  if ( c[J] == 0 ) // sender 0
  {
      *source = rank;
      
      rc[0] = c[0]; rc[1] = c[1]; rc[2] = c[2]; rc[3] = c[3];
      rc[J] = 1;

      Kernel_Coord2Rank( rc[0], rc[1], rc[2], rc[3], (unsigned *)target, &size );
      if ( (*target < 0) || (*target >= size) ) return 1;
  }

  if ( isTorus && (c[J] == P[J]/2) ) // sender P[J]/2
  {
      *source = rank;
      
      rc[0] = c[0]; rc[1] = c[1]; rc[2] = c[2]; rc[3] = c[3];
      rc[J] = P[J]/2 + 1;

      Kernel_Coord2Rank( rc[0], rc[1], rc[2], rc[3], (unsigned *)target, &size );
      if ( (*target < 0) || (*target >= size) ) return 1;
  }
  
  if ( c[J] == 1 ) // receiver 1
  {
      *target = rank;
      
      rc[0] = c[0]; rc[1] = c[1]; rc[2] = c[2]; rc[3] = c[3];
      rc[J] = 0;

      Kernel_Coord2Rank( rc[0], rc[1], rc[2], rc[3], (unsigned *)source, &size );
      if ( (*source < 0) || (*source >= size) ) return 2;
  }
  
  if ( isTorus && (c[J] == P[J]/2+1 ) ) // receiver P[J]/2+1
  {
      *target = rank;
      
      rc[0] = c[0]; rc[1] = c[1]; rc[2] = c[2]; rc[3] = c[3];
      rc[J] = P[J]/2;

      Kernel_Coord2Rank( rc[0], rc[1], rc[2], rc[3], (unsigned *)source, &size );
      if ( (*source < 0) || (*source >= size) ) return 2;
  }

  return 0;
}
