/* 
    Argonne Leadership Computing Facility benchmark
    BlueGene/P version
    Collectives, delivered latency
    Written by Vitali Morozov <morozov@anl.gov>

    Measures barrier, broadcast, and allreduce delivered latencies
        for MPI_COMM_WORLD, a copy of MPI_COMM_WORLD, and MPI_COMM_WORLD-1
*/    
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define MAXN 15000       // repetition rate for a single pair test
#define MAXSKIP 20       // skip first tests to warm-up
#define EXCLUDE_RANK 0   // the rank to exclude from World for World-1 benchmark

#define BSIZE 1024*8     // largest broadcast message size, bytes
#define RSIZE 256*8      // largest allreduce buffer size, bytes

unsigned long long getticks();
void do_barriers ( MPI_Comm comm, double *dt );
void do_broadcast( MPI_Comm comm, int Length, double *buf, double *dt ); 
void do_allreduce( MPI_Comm comm, int Length, double *sbuf, double *rbuf, double *dt ); 
int split( int c, MPI_Comm *half0, MPI_Comm *half1 );

static MPI_Comm comm[9];

struct benchmark
{
  const char *comm;
  double time[10];
};


main( int argc, char *argv[] )
{
    MPI_Group World_group, World_1_group;
    int ex_rank = EXCLUDE_RANK, i, taskid, ntasks;
    double bbuf[ BSIZE ], sbuf[ RSIZE ], rbuf[ RSIZE ], dtime[2];
    static struct benchmark b[] = 
    {
        { "MPI_COMM_WORLD", 0., 0. },
        { "Copy of MPI_COMM_WORLD", 0., 0. },
        { "MPI_COMM_WORLD-1", 0., 0. },
        { "xHalf0 of WORLD", 0., 0. },
        { "xHalf1 of WORLD", 0., 0. },
        { "yHalf0 of WORLD", 0., 0. },
        { "yHalf1 of WORLD", 0., 0. },
        { "zHalf0 of WORLD", 0., 0. },
        { "zHalf1 of WORLD", 0., 0. }
    };
    MPI_Init( &argc, &argv );
    
    comm[0] = MPI_COMM_WORLD;

    MPI_Comm_dup( MPI_COMM_WORLD, &comm[1] ); 

    MPI_Comm_group( MPI_COMM_WORLD, &World_group );
    MPI_Group_excl( World_group, 1, &ex_rank, &World_1_group );
    MPI_Comm_create( MPI_COMM_WORLD, World_1_group, &comm[2] );
    MPI_Group_free( &World_group );
    MPI_Group_free( &World_1_group );

    split( 0, &comm[3], &comm[4] );
    split( 1, &comm[5], &comm[6] );
    split( 2, &comm[7], &comm[8] );
    
    for ( i = 0; i < 9; i++ )
    {
        MPI_Barrier( MPI_COMM_WORLD );
        
        if ( comm[i] != MPI_COMM_NULL )
        {
            do_barriers ( comm[i], &b[i].time[0] );
            do_broadcast( comm[i],    2, bbuf, &b[i].time[2] );
            do_broadcast( comm[i], 1024, bbuf, &b[i].time[4] );
            do_allreduce( comm[i],    1, sbuf, rbuf, &b[i].time[6] );
            do_allreduce( comm[i],  256, sbuf, rbuf, &b[i].time[8] );
            
            MPI_Comm_rank( comm[i], &taskid );
            if ( taskid == 0 )
            {
                fprintf( stderr, "\n%20s:\n", b[i].comm );
                fprintf( stderr, "Barrier                 min, max, us: %18.6lf   %18.6lf\n", b[i].time[0], b[i].time[1] );
                fprintf( stderr, "Broadcast    2 doubles, min, max, us: %18.6lf   %18.6lf\n", b[i].time[2], b[i].time[3] );
                fprintf( stderr, "Broadcast 1024 doubles, min, max, us: %18.6lf   %18.6lf\n", b[i].time[4], b[i].time[5] );
                fprintf( stderr, "Allreduce    1 double , min, max, us: %18.6lf   %18.6lf\n", b[i].time[6], b[i].time[7] );
                fprintf( stderr, "Allreduce  256 double , min, max, us: %18.6lf   %18.6lf\n", b[i].time[8], b[i].time[9] );
            }
        }
    }

    MPI_Finalize();
}
    
void do_barriers( MPI_Comm comm, double *dt )    
{
    int k, taskid;
    double d;
    unsigned long long t1;

    for ( k = 0; k < MAXSKIP; k++ )
    {
        MPI_Barrier( comm );
    }
    
    t1 = getticks();
    for ( k = 0; k < MAXN; k++ )
    {
        MPI_Barrier( comm );
    }
    t1 = getticks() - t1;
    
    d = (double)t1 / (double)MAXN / (double)850.;  /* in us */
    
    MPI_Reduce( &d, &dt[0], 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, comm );
    MPI_Reduce( &d, &dt[1], 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, comm );

    return;
}

void do_broadcast( MPI_Comm comm, int Length, double *buf, double *dt )    
{
    int k, taskid;
    double d;
    unsigned long long t1;

    MPI_Barrier( comm );
    
    for ( k = 0; k < MAXSKIP; k++ )
    {
        MPI_Bcast( buf, Length, MPI_DOUBLE_PRECISION, 0, comm ); 
    }
    
    t1 = getticks();
    for ( k = 0; k < MAXN; k++ )
    {
        MPI_Bcast( buf, Length, MPI_DOUBLE_PRECISION, 0, comm ); 
    }
    t1 = getticks() - t1;
    
    d = (double)t1 / (double)MAXN / (double)850.;  /* in us */
    
    MPI_Reduce( &d, &dt[0], 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, comm );
    MPI_Reduce( &d, &dt[1], 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, comm );

    return;
}

void do_allreduce( MPI_Comm comm, int Length, double *sbuf, double *rbuf, double *dt )    
{
    int k, taskid;
    double d;
    unsigned long long t1;

    MPI_Barrier( comm );
    
    for ( k = 0; k < MAXSKIP; k++ )
    {
        MPI_Allreduce( sbuf, rbuf, Length, MPI_DOUBLE_PRECISION, MPI_SUM, comm ); 
    }
    
    t1 = getticks();
    for ( k = 0; k < MAXN; k++ )
    {
        MPI_Allreduce( sbuf, rbuf, Length, MPI_DOUBLE_PRECISION, MPI_SUM, comm ); 
    }
    t1 = getticks() - t1;
    
    d = (double)t1 / (double)MAXN / (double)850.;  /* in us */
    
    MPI_Reduce( &d, &dt[0], 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, comm );
    MPI_Reduce( &d, &dt[1], 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, comm );

    return;
}
